from pathlib import Path
from typing import Any, Type

from cg.constants.constants import FileFormat
from cg.constants.pacbio import PacBioDirsAndFiles
from cg.io.controller import ReadFile
from cg.services.pacbio.metrics.models import (
    ControlMetrics,
    HiFiMetrics,
    PolymeraseMetrics,
    ProductivityMetrics,
)
from cg.utils.files import get_file_in_directory


class MetricsParser:
    """Class for parsing PacBio sequencing metrics."""

    def __init__(self, smrt_cell_path: Path) -> None:
        self.smrt_cell_path: Path = smrt_cell_path
        self.report_dir = Path(smrt_cell_path, "statistics")
        # For HiFi metrics
        self.base_calling_report_file: Path = get_file_in_directory(
            directory=self.report_dir, file_name=PacBioDirsAndFiles.BASECALLING_REPORT
        )
        self.hifi_metrics: HiFiMetrics = self.parse_attributes_to_model(
            report_file=self.base_calling_report_file, data_model=HiFiMetrics
        )
        # For control metrics
        self.control_report_file: Path = get_file_in_directory(
            directory=self.report_dir, file_name=PacBioDirsAndFiles.CONTROL_REPORT
        )
        self.control_metrics: ControlMetrics = self.parse_attributes_to_model(
            report_file=self.control_report_file, data_model=ControlMetrics
        )
        # For productivity metrics
        self.loading_report_file: Path = get_file_in_directory(
            directory=self.report_dir, file_name=PacBioDirsAndFiles.LOADING_REPORT
        )
        self.productivity_metrics: ProductivityMetrics = self.parse_attributes_to_model(
            report_file=self.loading_report_file, data_model=ProductivityMetrics
        )
        # For polymerase metrics
        self.raw_data_report_file: Path = get_file_in_directory(
            directory=self.report_dir, file_name=PacBioDirsAndFiles.RAW_DATA_REPORT
        )
        self.polymerase_metrics: PolymeraseMetrics = self.parse_attributes_to_model(
            report_file=self.raw_data_report_file, data_model=PolymeraseMetrics
        )

    @staticmethod
    def _parse_report(report_file: Path) -> list[dict[str, Any]]:
        """Parse the attribute element of a PacBio report file in JSON format."""
        parsed_json: dict = ReadFile.read_file[FileFormat.JSON](file_path=report_file)
        return parsed_json.get("attributes")

    def parse_attributes_to_model(
        self,
        report_file: Path,
        data_model: Type[ControlMetrics | HiFiMetrics | PolymeraseMetrics | ProductivityMetrics],
    ) -> ControlMetrics | HiFiMetrics | PolymeraseMetrics | ProductivityMetrics:
        """Parse the attributes to a model."""
        report_content: list[dict[str, Any]] = self._parse_report(report_file=report_file)
        data: dict = {report_field["id"]: report_field["value"] for report_field in report_content}
        return data_model.model_validate(data, from_attributes=True)



from pydantic import BaseModel, Field, field_validator

from cg.constants.pacbio import (
    CCSAttributeIDs,
    ControlAttributeIDs,
    LoadingAttributesIDs,
    PolymeraseDataAttributeIDs,
)
from cg.utils.calculations import divide_by_thousand_with_one_decimal, fraction_to_percent


class HiFiMetrics(BaseModel):
    """Model for the HiFi metrics."""

    reads: int = Field(..., alias=CCSAttributeIDs.NUMBER_OF_READS)
    yield_: int = Field(..., alias=CCSAttributeIDs.TOTAL_NUMBER_OF_BASES)
    mean_read_length_kb: float = Field(..., alias=CCSAttributeIDs.MEAN_READ_LENGTH)
    median_read_length: int = Field(..., alias=CCSAttributeIDs.MEDIAN_READ_LENGTH)
    mean_length_n50: int = Field(..., alias=CCSAttributeIDs.READ_LENGTH_N50)
    median_read_quality: str = Field(..., alias=CCSAttributeIDs.MEDIAN_ACCURACY)
    percent_q30: float = Field(..., alias=CCSAttributeIDs.PERCENT_Q30)

    _validate_mean_read_length_kb = field_validator("mean_read_length_kb", mode="before")(
        divide_by_thousand_with_one_decimal
    )
    _validate_percent_q30 = field_validator("percent_q30", mode="before")(fraction_to_percent)


class ControlMetrics(BaseModel):
    """Model for the control metrics."""

    reads: int = Field(..., alias=ControlAttributeIDs.NUMBER_OF_READS)
    mean_read_length_kb: float = Field(..., alias=ControlAttributeIDs.MEAN_READ_LENGTH)
    percent_mean_concordance_reads: float = Field(
        ..., alias=ControlAttributeIDs.PERCENT_MEAN_READ_CONCORDANCE
    )
    percent_mode_concordance_reads: float = Field(
        ..., alias=ControlAttributeIDs.PERCENT_MODE_READ_CONCORDANCE
    )

    _validate_mean_read_length_kb = field_validator("mean_read_length_kb", mode="before")(
        divide_by_thousand_with_one_decimal
    )
    _validate_percent_mean_concordance_reads = field_validator(
        "percent_mean_concordance_reads", mode="before"
    )(fraction_to_percent)

    _validate_percent_mode_concordance_reads = field_validator(
        "percent_mode_concordance_reads", mode="before"
    )(fraction_to_percent)


class ProductivityMetrics(BaseModel):
    """Model for the loading metrics."""

    productive_zmws: int = Field(..., alias=LoadingAttributesIDs.PRODUCTIVE_ZMWS)
    p_0: int = Field(..., alias=LoadingAttributesIDs.P_0)
    p_1: int = Field(..., alias=LoadingAttributesIDs.P_1)
    p_2: int = Field(..., alias=LoadingAttributesIDs.P_2)

    @property
    def percentage_p_0(self) -> float:
        return self._calculate_percentage(self.p_0)

    @property
    def percentage_p_1(self) -> float:
        return self._calculate_percentage(self.p_1)

    @property
    def percentage_p_2(self) -> float:
        return self._calculate_percentage(self.p_2)

    def _calculate_percentage(self, value: int) -> float:
        """Calculates the percentage of a value to productive_zmws."""
        if self.productive_zmws == 0:
            return 0.0
        return round((value / self.productive_zmws) * 100, 0)


class PolymeraseMetrics(BaseModel):
    """Model for the polymerase metrics."""

    mean_read_length: float = Field(..., alias=PolymeraseDataAttributeIDs.MEAN_READ_LENGTH)
    read_length_n50: float = Field(..., alias=PolymeraseDataAttributeIDs.READ_LENGTH_N50)
    mean_longest_subread_length: float = Field(
        ..., alias=PolymeraseDataAttributeIDs.MEAN_LONGEST_SUBREAD_LENGTH
    )
    longest_subread_length_n50: float = Field(
        ..., alias=PolymeraseDataAttributeIDs.LONGEST_SUBREAD_LENGTH_N50
    )

    _validate_mean_read_length = field_validator("mean_read_length", mode="before")(
        divide_by_thousand_with_one_decimal
    )
    _validate_read_length_n50 = field_validator("read_length_n50", mode="before")(
        divide_by_thousand_with_one_decimal
    )
    _validate_mean_longest_subread_length = field_validator(
        "mean_longest_subread_length", mode="before"
    )(divide_by_thousand_with_one_decimal)
    _validate_longest_subread_length_n50 = field_validator(
        "longest_subread_length_n50", mode="before"
    )(divide_by_thousand_with_one_decimal)



import math
from pathlib import Path
from typing import Any, Type

import pytest
from _pytest.fixtures import FixtureRequest

from cg.constants.pacbio import CCSAttributeIDs, ControlAttributeIDs
from cg.services.pacbio.metrics.metrics_parser import MetricsParser
from cg.services.pacbio.metrics.models import (
    ControlMetrics,
    HiFiMetrics,
    PolymeraseMetrics,
    ProductivityMetrics,
)


def test_metrics_parser_initialisation(pac_bio_smrt_cell_dir: Path):
    """Test the initialisation of the metrics parser."""
    # GIVEN a PacBio SMRT cell path

    # WHEN initialising the metrics parser
    parser = MetricsParser(smrt_cell_path=pac_bio_smrt_cell_dir)

    # THEN assert that the parser is initialised with the expected attributes
    assert isinstance(parser.hifi_metrics, HiFiMetrics)
    assert isinstance(parser.control_metrics, ControlMetrics)
    assert isinstance(parser.productivity_metrics, ProductivityMetrics)
    assert isinstance(parser.polymerase_metrics, PolymeraseMetrics)


@pytest.mark.parametrize(
    "report_file_path",
    [
        "pac_bio_control_report",
        "pac_bio_css_report",
        "pac_bio_loading_report",
        "pac_bio_raw_data_report",
    ],
)
def test_parse_attributes_from_json(
    pac_bio_metrics_parser: MetricsParser,
    report_file_path: str,
    request: FixtureRequest,
):
    """Test the parsing of attributes from any PacBio report file."""
    # GIVEN a PacBio report file and a PacBio metrics parser initialised from the SMRTcell path
    report_file: Path = request.getfixturevalue(report_file_path)

    # WHEN parsing the report file
    attributes: list[dict[str, Any]] = pac_bio_metrics_parser._parse_report(report_file=report_file)

    # THEN assert that the report attributes are parsed correctly
    assert isinstance(attributes, list)
    assert isinstance(attributes[0], dict)
    assert "id" in attributes[0]
    assert "value" in attributes[0]


@pytest.mark.parametrize(
    "report_file_path, model, metrics_fixture, percent_fields",
    [
        (
            "pac_bio_control_report",
            ControlMetrics,
            "pac_bio_control_metrics",
            [
                ControlAttributeIDs.PERCENT_MEAN_READ_CONCORDANCE,
                ControlAttributeIDs.PERCENT_MODE_READ_CONCORDANCE,
            ],
        ),
        ("pac_bio_css_report", HiFiMetrics, "pac_bio_hifi_metrics", [CCSAttributeIDs.PERCENT_Q30]),
        ("pac_bio_raw_data_report", PolymeraseMetrics, "pac_bio_polymerase_metrics", []),
        ("pac_bio_loading_report", ProductivityMetrics, "pac_bio_productivity_metrics", []),
    ],
    ids=["Control", "Hi-Fi", "Polymerase", "Productivity"],
)
def test_parse_attributes_to_model(
    pac_bio_metrics_parser: MetricsParser,
    report_file_path: str,
    model: Type[ControlMetrics | HiFiMetrics | ProductivityMetrics],
    metrics_fixture: str,
    percent_fields: list[str],
    request: FixtureRequest,
):
    """Test to parse the attributes to a metrics model."""
    # GIVEN a metrics parser

    # GIVEN a pac-bio report file
    report_file: Path = request.getfixturevalue(report_file_path)

    # GIVEN a metrics object with the expected parsed metrics
    expected_metrics: ControlMetrics | HiFiMetrics | ProductivityMetrics = request.getfixturevalue(
        metrics_fixture
    )

    # WHEN parsing the attributes to a given metrics model
    parsed_metrics: ControlMetrics | HiFiMetrics | ProductivityMetrics = (
        pac_bio_metrics_parser.parse_attributes_to_model(
            report_file=report_file,
            data_model=model,
        )
    )

    # THEN assert that the model attributes are the expected ones
    assert parsed_metrics == expected_metrics

    # THEN assert that the percentage fields of the model are not taken as a fraction
    metrics_dict: dict = parsed_metrics.dict(by_alias=True)
    for percent_field in percent_fields:
        assert metrics_dict.get(percent_field) > 1


def test_productivity_metrics_percentage_attributes(
    pac_bio_productivity_metrics: ProductivityMetrics,
):
    """Test the percentage attributes of the productivity metrics."""
    # GIVEN a productivity metrics object

    # WHEN accessing the percentage attributes
    percentage_p_0: float = pac_bio_productivity_metrics.percentage_p_0
    percentage_p_1: float = pac_bio_productivity_metrics.percentage_p_1
    percentage_p_2: float = pac_bio_productivity_metrics.percentage_p_2

    # THEN assert that the percentage attributes are calculated correctly
    assert math.isclose(percentage_p_0, 40, abs_tol=1e-9)
    assert math.isclose(percentage_p_1, 60, abs_tol=1e-9)
    assert math.isclose(percentage_p_2, 0, abs_tol=1e-9)



"""Module for PacBio fixtures returning strings."""

import pytest


@pytest.fixture
def pac_bio_test_run_name() -> str:
    """Return the name of a PacBio SMRT cell."""
    return "r84202_20240522_133539"



"""Module for PacBio fixtures returning service objects."""

from pathlib import Path

import pytest

from cg.services.pacbio.metrics.metrics_parser import MetricsParser


@pytest.fixture
def pac_bio_metrics_parser(pac_bio_smrt_cell_dir: Path) -> MetricsParser:
    """Return a PacBio metrics parser."""
    return MetricsParser(pac_bio_smrt_cell_dir)



"""Module for PacBio fixtures returning Path objects."""

from pathlib import Path

import pytest

from cg.constants.pacbio import PacBioDirsAndFiles


# Directory fixtures
@pytest.fixture
def pac_bio_fixtures_dir(devices_dir: Path) -> Path:
    """Return the path to the PacBio fixtures directory."""
    return Path(devices_dir, "pacbio")


@pytest.fixture
def pac_bio_runs_dir(pac_bio_fixtures_dir: Path) -> Path:
    """Return the path to the PacBio run directory."""
    return Path(pac_bio_fixtures_dir, "SMRTcells")


@pytest.fixture
def pac_bio_test_run_dir(pac_bio_runs_dir: Path, pac_bio_test_run_name: str) -> Path:
    """Return the path to a PacBio run directory."""
    return Path(pac_bio_runs_dir, pac_bio_test_run_name)


@pytest.fixture
def pac_bio_smrt_cell_dir(pac_bio_test_run_dir: Path) -> Path:
    """Return the path to a PacBio SMRT cell directory."""
    return Path(pac_bio_test_run_dir, "1_A01")


@pytest.fixture
def pac_bio_run_statistics_dir(pac_bio_smrt_cell_dir: Path) -> Path:
    """Return the path to the PacBio SMRT cell statistics directory."""
    return Path(pac_bio_smrt_cell_dir, "statistics")


# File fixtures


@pytest.fixture
def pac_bio_css_report(pac_bio_run_statistics_dir: Path) -> Path:
    """Return the path to the PacBio CSS report."""
    return Path(pac_bio_run_statistics_dir, PacBioDirsAndFiles.BASECALLING_REPORT)


@pytest.fixture
def pac_bio_control_report(pac_bio_run_statistics_dir: Path) -> Path:
    """Return the path to the PacBio control report."""
    return Path(pac_bio_run_statistics_dir, PacBioDirsAndFiles.CONTROL_REPORT)


@pytest.fixture
def pac_bio_loading_report(pac_bio_run_statistics_dir: Path) -> Path:
    """Return the path to the PacBio loading report."""
    return Path(pac_bio_run_statistics_dir, PacBioDirsAndFiles.LOADING_REPORT)


@pytest.fixture
def pac_bio_raw_data_report(pac_bio_run_statistics_dir: Path) -> Path:
    """Return the path to the PacBio raw data report."""
    return Path(pac_bio_run_statistics_dir, PacBioDirsAndFiles.RAW_DATA_REPORT)



"""CLI support to create config and/or start Tomte."""

import logging

import click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.cli.workflow.commands import resolve_compression
from cg.cli.workflow.nf_analysis import (
    config_case,
    metrics_deliver,
    report_deliver,
    run,
    start,
    start_available,
    store,
    store_available,
    store_housekeeper,
)
from cg.constants.constants import MetaApis
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.tomte import TomteAnalysisAPI

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True, context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_context
def tomte(context: click.Context) -> None:
    """gms/tomte analysis workflow."""
    AnalysisAPI.get_help(context)
    context.obj.meta_apis[MetaApis.ANALYSIS_API] = TomteAnalysisAPI(config=context.obj)


tomte.add_command(resolve_compression)
tomte.add_command(config_case)
tomte.add_command(run)
tomte.add_command(start)
tomte.add_command(start_available)
tomte.add_command(metrics_deliver)
tomte.add_command(report_deliver)
tomte.add_command(store_housekeeper)
tomte.add_command(store)
tomte.add_command(store_available)
